# Standard Library Imports
from abc import ABC, abstractmethod

# Local Imports
from ..solvers.solver import Solver
from ..utils.game_status import GameStatus
from ..utils.question_outcome import QuestionOutcome


class Strategy(ABC):
    """An abstract base class for solving questions base on the current game
    status.

    A `Strategy` answers questions getting an answer from an appropriate
    `Solver` based on the current game status.
    """

    def __init__(self) -> None:
        """Initializes the strategy."""
        return

    @abstractmethod
    def get_solver(self, game_variables: dict, status: GameStatus) -> Solver:
        """Chooses the appropriate solver based on the game variables and
        current game status.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.
            status (GameStatus): The current game status (e.g. number of lives)

        Returns:
            Solver: The choosen solver to answer the question.
        """
        pass

    def solve(self, game_variables: dict, status: GameStatus) -> tuple[str, QuestionOutcome]:
        """Solves the question by getting an answer from the appropriate solver.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.
            status (GameStatus): The current game status (e.g. number of lives)

        Returns:
            tuple[str, QuestionOutcome]: A tuple of the chosen answer and the corresponding outcome.
        """
        solver = self.get_solver(game_variables, status)
        return solver.solve(game_variables)



from .observer import Observer
from .periodic_observer import PeriodicObserver
from .progress_bar_observer import ProgressBarObserver
from .status_observer import StatusObserver
from .timeout_observer import TimeOutObserver

__all__ = ["Observer", "PeriodicObserver", "ProgressBarObserver", "StatusObserver", "TimeOutObserver"]



# Standard Library Imports
import re

# Local Imports
from .json_file_logger import JsonFileLogger


class WordData:
    """A class to process and store word-related data extracted from game
    variables.

    This class extracts and filters relevant word readings and meanings from
    the provided game data.
    """

    def __init__(self, game_variables: dict) -> None:
        """Initializes the WordData.

        Args:
            game_variables (dict): The game variables shortly after a solution was revealed.
        """
        data = game_variables["_data"]

        self.index: str = str(data[6])
        self.readings: list[str] = [self._filter(reading) for reading in data[9:12]]
        self.readings = [
            reading for reading in self.readings if self._is_reading(reading)
        ]
        self.meanings: list[str] = [self._filter(meaning) for meaning in data[19:21]]
        self.meanings = [
            meaning for meaning in self.meanings if self._is_meaning(meaning)
        ]

    @classmethod
    def _filter(cls, string: str) -> str:
        """Removes unwanted characters from a string.

        Args:
            string (str): The string to be filtered.

        Returns:
            str: The filtered string with unwanted characters removed.
        """
        string = string.replace("₨", "")
        string = string.replace("\x1bI[48]", "𥝱")
        string = re.sub(r"\x1bI\[.*?]", "", string)
        string = re.sub(r"\x1bc\[.*?]", "", string)
        return string

    @classmethod
    def _is_reading(cls, reading: str) -> bool:
        """Checks if the given reading is not a dummy.

        Args:
            reading (str): The reading to be checked.

        Returns:
            bool: True if the reading is not a dummy, False if it is a dummy.
        """
        return reading != "000000000000000000000"

    @classmethod
    def _is_meaning(cls, meaning: str) -> bool:
        """Checks if the given meaning is not a dummy.

        Args:
            meaning (str): The meaning to be checked.

        Returns:
            bool: True if the meaning is not a dummy, False if it is a dummy.
        """
        return meaning != "\u3000"


class SingleLevelWordsLogger(JsonFileLogger):
    """A `Logger` logs data for a single level into a JSON file.

    play through rounds by answering questions using the specified `Strategy`.
    This `Logger` is a `JsonFileLogger` that stores readings and meanings of
    question coming from a single level.

    Attributes:
        questions_for_levels (dict[int, int]): The number of questions for the levels in 漢字でGO!.
    """

    questions_for_levels: dict[int, int] = {
        1: 510,
        2: 760,
        3: 1010,
        4: 1420,
        5: 1810,
        6: 1160,
        7: 500,
    }

    def __init__(self, level: int, path: str) -> None:
        """Initializes the logger.

        Args:
            level (int): The level for which words should be logged.
            path (str): The path to the JSON file where log data will be stored.
        """
        super().__init__(path)
        if not bool(self.data):
            self.data: dict = {
                str(index): None
                for index in range(1, self.questions_for_levels[level] + 1)
            }

    def log(self, game_variables: dict) -> None:
        """Logs the readings and meanings from the given game variables.

        Args:
            game_variables (dict): The game variables shortly after a solution was revealed.
        """
        word_data = WordData(game_variables)
        data_as_dict = {"readings": word_data.readings, "meanings": word_data.meanings}
        if self.data[word_data.index] != data_as_dict:
            self.data[word_data.index] = data_as_dict
            self._write_data()



# Local Imports
from .logger import Logger


class BaseLogger(Logger):
    """A basic `Logger` that does not do actual logging"""

    def __init__(self) -> None:
        """Initializes the logger."""
        return

    def log(self, game_variables: dict) -> None:
        """Does not do anything.

        Args:
            game_variables (dict): The game variables shortly after a solution was revealed.
        """
        return



from .correct_solver import CorrectSolver
from .quit_solver import QuitSolver
from .skip_solver import SkipSolver
from .solver import Solver

__all__ = ["CorrectSolver", "QuitSolver", "SkipSolver", "Solver"]



# Local Imports
from .solver import Solver
from ..utils.question_outcome import QuestionOutcome


class QuitSolver(Solver):
    """A `Solver` that always chooses to quit."""

    def solve(self, game_variables: dict) -> tuple[str, QuestionOutcome]:
        """Chooses to quit the question.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.

        Returns:
            tuple[str, QuestionOutcome]: A tuple of 'q' and the 'Incorrect' outcome.
        """
        return "q", QuestionOutcome(False, False)



# Standard Library Imports
import asyncio

# Third-Party Imports
from playwright.async_api import Page

# Local Imports
from .playwright import press_key, press_key_sequence
from .speed import Speed
from ..world_maps.world_map import Popups, WorldMap


class MenuNavigator:
    """A class for navigating through the game.

    A `MenuNavigator` can navigate to a given mode and level while keeping
    track of the current position given a `WorldMap`. Additionally, it can
    set the number of lives and questions and also keep track of them.
    """

    def __init__(self, kadego: Page, world_map: WorldMap, speed: float) -> None:
        """Initializes the navigator.

        Args:
            kadego (Page): The Playwright page for playing 漢字でGO!.
            world_map (WorldMap): The world map to use.
            speed (float): The speed multiplier for bot actions.
        """
        self.kadego: Page = kadego
        self.world_map: WorldMap = world_map
        self.speed: float = speed

        self.started: bool = False
        self.level_positions: dict[str, str] = {mode: world_map.get_levels(mode)[0] for mode in world_map.get_modes()}
        self.settings_positions: list[int] = [0, 0, 0]
        self.mode: str | None = None
        self.level: str | None = None
        self.questions: int = 16
        self.settings: int = 0

    async def _wait(self, delay: float) -> None:
        """Pauses for a specified amount of time.
        speed.

        Args:
            delay (float): The time to wait in seconds.
        """
        await asyncio.sleep(delay * self.speed)

    async def _shift(self, amount: int) -> None:
        """Shifts to the right (left) by the specified amount.

        Args:
            amount (float): The amount to be shifted by, (+) -> right, (-) -> left.
        """
        if amount < 0:
            await press_key_sequence(self.kadego, amount * ["ArrowLeft"], Speed.SHORT.value)
        else:
            await press_key_sequence(self.kadego, amount * ["ArrowRight"], Speed.SHORT.value)
        await self._wait(Speed.SHORT.value)

    async def _click_through_popups(self, popups: Popups) -> None:
        """Clicks through the specified popups.

        Args:
            popups (Popups): The list of button presses necessary to close the popups.
        """
        await press_key_sequence(self.kadego, popups, Speed.MEDIUM.value)
        if bool(popups):
            await self._wait(Speed.LONG.value)

    async def _start_game(self) -> None:
        """Starts the game and clicks through the popups"""
        self.status = "Waiting for game to start"
        selector = "#_111_input"
        await self.kadego.wait_for_selector(selector)
        await self.kadego.press(selector, "Enter")
        await self._wait(Speed.LONG.value)

        popups = self.world_map.get_popups()
        await self._click_through_popups(popups)
        self.world_map.clear_popups()
        self.started = True

    async def _enter_mode(self) -> None:
        """Enters the current mode and clicks through the popups."""
        await press_key(self.kadego, "Space")
        await self._wait(Speed.MEDIUM.value)
        self.level = self.level_positions[self.mode]

        popups = self.world_map.get_mode_popups(self.mode)
        await self._click_through_popups(popups)
        self.world_map.clear_mode_popups(self.mode)

    async def _set_lives(self, lives: int) -> None:
        """Sets the number of lives to the specified amount.

        Args:
            lives (int): The number of lives to set.
        """
        await press_key(self.kadego, "ArrowDown")
        await self._wait(Speed.MEDIUM.value)
        down = ((2 - self.settings) % 3) * ["ArrowDown"]
        right = ((lives - 1 - self.settings_positions[2]) % 5) * ["ArrowRight"]
        keys = down + right + ["Enter", "Shift"]
        await press_key_sequence(self.kadego, keys, Speed.MEDIUM.value)
        await self._wait(Speed.MEDIUM.value)
        self.settings_positions[2] = lives - 1
        self.settings = 2

    async def _enter_level(self) -> None:
        """Enters the current level and clicks through the popups."""
        await press_key(self.kadego, "Space")
        await self._wait(Speed.MEDIUM.value)

        popups = self.world_map.get_mode_level_popups(self.mode, self.level)
        await self._click_through_popups(popups)
        self.world_map.clear_mode_level_popups(self.mode, self.level)

    async def _set_questions(self, questions: int) -> None:
        """Sets the number of questions to the specified amount.

        Args:
            questions (int): The number of questions to set.
        """
        questions_map = {7: 2, 10: 1, 16: 0}
        position_own = questions_map[self.questions]
        position = questions_map[questions]
        keys = ((position - position_own) % 3) * ["ArrowUp"]
        await press_key_sequence(self.kadego, keys, Speed.SHORT.value)
        await self._wait(Speed.MEDIUM.value)
        self.questions = questions

    async def _start_round(self) -> None:
        """Starts the round."""
        start = self.world_map.get_round_start(self.mode, self.level)
        await press_key_sequence(self.kadego, start, Speed.SHORT.value)

    async def navigate_to_mode(self, mode: str) -> None:
        """Navigates to the specified mode.

        Args:
            mode (str): The mode to navigate to.
        """
        if not self.started:
            await self._start_game()
        if self.level is not None:
            await press_key(self.kadego, "Shift")
            await self._wait(Speed.SHORT.value)
            self.level = None

        position_own = 0 if self.mode is None else self.world_map.get_mode_direction(self.mode)
        position = self.world_map.get_mode_direction(mode)
        await self._shift(position - position_own)
        self.mode = mode

    async def navigate_to_level(self, mode: str, level: str) -> None:
        """Navigates to the specified level of a mode.

        Args:
            mode (str): The mode of the given level.
            level (str): The level to navigate to.
        """
        if self.mode != mode:
            await self.navigate_to_mode(mode)
        if self.level is None:
            await self._enter_mode()

        position_own = self.world_map.get_level_direction(self.mode, self.level)
        position = self.world_map.get_level_direction(mode, level)
        await self._shift(position - position_own)
        self.level_positions[mode] = level
        self.level = level

    async def start_round(self, mode: str, level: str, lives: int, questions: int) -> None:
        """Starts a round with the given specifications.

        Args:
            mode (str): The mode to play.
            level (str): The level of the mode to play.
            lives (int): The number of lives to set.
            questions (int): The number of questions to set.
        """
        await self.navigate_to_level(mode, level)
        await self._set_lives(lives)
        await self._enter_level()
        await self._set_questions(questions)
        await self._start_round()

    async def reload(self) -> None:
        """Reloads the game."""
        await self.kadego.reload()
        self.started = False
        self.level = None



# Local Imports
from .solver import Solver
from ..utils.question_outcome import QuestionOutcome


class SkipSolver(Solver):
    """A `Solver` that always chooses to skip."""

    def solve(self, game_variables: dict) -> tuple[str, QuestionOutcome]:
        """Chooses to skip the question.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.

        Returns:
            tuple[str, QuestionOutcome]: A tuple of an empty string and the 'Skipped' outcome.
        """
        return "", QuestionOutcome(True, True)



from .utils.game_status import GameStatus
from .utils.menu_navigator import MenuNavigator
from .utils.question_outcome import QuestionOutcome
from .utils.playwright import get_kadego

__all__ = ["GameStatus", "MenuNavigator", "QuestionOutcome", "get_kadego"]



# Local Imports
from .world_map import WorldMap

MODES = {"main": 1, "casual": 2, "rush": 3, "extra": 4}

MAIN_LEVELS = {"normal": 0, "hard": 1, "very_hard": 2, "hell": 3}
CASUAL_LEVELS = {
    "basic": 0,
    "variety": 1,
    "variety_hard": 2,
    "one_character": 3,
    "four_characters": 4,
    "places": 5,
    "bugs": 6,
    "mammals": 7,
    "birds": 8,
    "aquatic": 9,
    "plants": 10,
    "food": 11,
}
RUSH_LEVELS = {
    "level_1": 0,
    "level_2": 1,
    "level_3": 2,
    "level_4": 3,
    "level_5": 4,
    "level_6": 5,
    "level_7": 5,
}
EXTRA_LEVELS = {"math": 0, "math_hard": 1, "english": 2, "elements": 3}
MODE_LEVELS = {
    "main": MAIN_LEVELS,
    "casual": CASUAL_LEVELS,
    "rush": RUSH_LEVELS,
    "extra": EXTRA_LEVELS,
}

POPUPS = ["Space", "Shift", "Space", "Shift", "Space"]
MODE_POPUPS = {"main": ["Space"]}
MODE_LEVEL_POPUPS = {"main": {"hell": ["Space"]}}

ROUND_STARTS = {"rush": {"level_7": ["ArrowRight+Space"]}}


class KanjiDeGoMap(WorldMap):
    """A `WorldMap` of 漢字でGO!."""

    def __init__(self) -> None:
        """Initializes the WorldMap of 漢字でGO!"""
        super().__init__(
            MODES, MODE_LEVELS, POPUPS, MODE_POPUPS, MODE_LEVEL_POPUPS, ROUND_STARTS
        )



# Standard Library Imports
from dataclasses import dataclass


@dataclass
class QuestionOutcome:
    """A class representing the outcome of a question being answered.

    Attributes:
        correct (bool): Indicates whether the question was answered correctly.
        skipped (bool): Indicates whether the question was skipped.
    """

    correct: bool
    skipped: bool



# Standard Library Imports
from abc import ABC, abstractmethod

# Local Imports
from ..utils.question_outcome import QuestionOutcome


class Solver(ABC):
    """ An abstract base class for solving a question.

    A `Solver` chooses an answer and a corresponding `QuestionOutcome`
    given the game variables.
    """

    def __init__(self) -> None:
        """Initializes the solver."""
        return

    @abstractmethod
    def solve(self, game_variables: dict) -> tuple[str, QuestionOutcome]:
        """Chooses an answer and an outcome based on provided game variables.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.

        Returns:
            tuple[str, QuestionOutcome]: A tuple of the chosen answer and the corresponding outcome.
        """
        pass



from .base_logger import BaseLogger
from .json_file_logger import JsonFileLogger
from .logger import Logger
from .single_level_words_logger import SingleLevelWordsLogger

__all__ = ["BaseLogger", "JsonFileLogger", "Logger", "SingleLevelWordsLogger"]



# Local Imports
from .strategy import Strategy
from ..solvers.correct_solver import CorrectSolver
from ..solvers.quit_solver import QuitSolver
from ..solvers.skip_solver import SkipSolver
from ..solvers.solver import Solver
from ..utils.game_status import GameStatus


class CuriousStrategy(Strategy):
    """A `Strategy` aims to maximize the amount of questions in a round."""

    def get_solver(self, game_variables: dict, status: GameStatus) -> Solver:
        """Selects to skip when possible and answer correctly otherwise. That
        is except for at the peniultimate question where it will select to
        quit until only one life remaining and on the last question where it
        will also choose to quit after skippingis no longer possible.

        Args:
             game_variables (dict): The game variables shortly after a question was posed.
             status (GameStatus): The current game status (e.g. number of lives)

         Returns:
             tuple[str, QuestionOutcome]: A tuple of the chosen answer and the corresponding outcome.
        """
        if status.skip:
            return SkipSolver()
        if status.question == status.questions - 1 and status.lives > 1:
            return QuitSolver()
        if status.question == status.questions and not status.skip:
            return QuitSolver()
        return CorrectSolver()



# Standard Library Imports
from dataclasses import dataclass

# Local Imports
from ..utils.question_outcome import QuestionOutcome


@dataclass
class GameStatus:
    """A class representing the current status of the game and updates
    when given a `QuestionOutcome`.

    Attributes:
        lives (int): The number of lives remaining.
        questions (int): The total number of questions.
        question (int): The current question number.
        skip (bool): Flag indicating whether skipping is possible.
    """

    lives: int
    questions: int
    question: int = 1
    skip: bool = True

    def update(self, outcome: QuestionOutcome) -> None:
        """Updates the game status based on the outcome of a question.

        Args:
            outcome (QuestionOutcome): The outcome of the question
        """
        if outcome.skipped:
            self.skip = False
        elif outcome.correct:
            if self.question % (self.questions // 3) == 0:
                self.skip = True
            self.question += 1
        else:
            self.lives -= 1

    def is_done(self) -> bool:
        """Checks if the round is over.

        Returns:
            bool: True if not lives remain or the total number of questions was reached, False otherwise.
        """
        return self.lives == 0 or self.question > self.questions

    def is_win(self) -> bool:
        """Checks if the round was one.

        Returns:
            bool: True if the round is done and lives still remain, False otherwise.
        """
        return self.is_done() and self.lives > 0

    def is_loss(self) -> bool:
        """Checks if the round was one.

        Returns:
            bool: True if the round is done and not lives remain, False otherwise.
        """
        return self.is_done() and self.lives == 0



# Standard Library Imports
import asyncio
import time

# Local Imports
from .periodic_observer import PeriodicObserver
from ..bots import Bot


class TimeOutObserver(PeriodicObserver):
    """An `Observer` that stops the `Bot` from being run after a specified
    amount of time.

    This `Observer` periodically checks how long a `Bot` has been running for
    already. If this time exceeds some specified threshold the `Runner` will
    be stopped from running the `Bot`.
    """

    def __init__(self, bot: Bot, timeout: float, delay: float = 0.1) -> None:
        """Initializes the observer given a bot and delay.

        Args:
            bot (Bot): The bot to be observed.
            timeout (float): The time (in seconds) after which the bot will be stopped from being run.
            delay (float, optional): The delay (in seconds) between each observation cycle (default is 0.2).
        """
        super().__init__(bot, delay)
        self.timeout = timeout
        self.end_time: float | None = None

    async def _loop_function(self, stop_event: asyncio.Event) -> None:
        """Stops the bot from being run if the timeout was exceeded.

        Args:
            stop_event (asyncio.Event): Event that when set stops the runner from running the bot.
        """
        if self.end_time is None:
            self.end_time = time.time() + self.timeout
        if time.time() > self.end_time:
            stop_event.set()



from .always_correct_strategy import AlwaysCorrectStrategy
from .always_quit_strategy import AlwaysQuitStrategy
from .curious_strategy import CuriousStrategy
from .strategy import Strategy

__all__ = ["AlwaysCorrectStrategy", "AlwaysQuitStrategy", "CuriousStrategy", "Strategy"]



from enum import Enum


class Speed(Enum):
    """An enumeration representing different durations for actions.

    Attributes:
        SHORT (float, optinal): A short duration (default is 0.1).
        MEDIUM (float, optional): A medium duration (default is 1.0).
        LONG (float, optional): A long duration (default is 10.0).
    """
    SHORT: float = 0.1
    MEDIUM: float = 1.0
    LONG: float = 10.0



# Standard Library Imports
import asyncio

# Third-Party Imports
from playwright.async_api import Page

# Local Imports
from ..exceptions.invalid_bot_configuration_error import InvalidBotConfigurationError
from ..loggers.base_logger import BaseLogger
from ..loggers.logger import Logger
from ..strategies.always_correct_strategy import AlwaysCorrectStrategy
from ..strategies.strategy import Strategy
from ..utils.game_status import GameStatus
from ..utils.menu_navigator import MenuNavigator
from ..utils.playwright import get_game_variables, press_key, press_key_sequence
from ..utils.question_outcome import QuestionOutcome
from ..utils.speed import Speed
from ..world_maps.kanji_de_go_map import KanjiDeGoMap
from ..world_maps.world_map import WorldMap


class Bot:
    """A class for automating the process of playing 漢字でGO!.

    A `Bot` can navigate to a given mode and level using a given `WorldMap`
    and set the number of lives and questions accordingly. It can also then
    play through rounds by answering questions using the specified `Strategy`.
    Additonally, question data can be logged by a given `Logger`.
    """

    def __init__(
        self,
        kadego: Page,
        mode: str,
        level: str,
        strategy: Strategy,
        lives: int = 5,
        questions: int = 16,
        logger: Logger | None = None,
        world_map: WorldMap | None = None,
        speed: float = 1.0,
    ) -> None:
        """Initializes the bot with necessary parameters.

        Args:
            kadego (Page): The Playwright page for playing 漢字でGO!.
            mode (str): The mode to play.
            level (str): The level to play.
            strategy (Strategy): The strategy for answering questions.
            lives (int, optional): The number of lives to set.
            questions (int, optional): The number of questions to set (default is 16).
            logger (Logger, optional): The logger to use (default is `BaseLogger`).
            world_map (WorldMap, optional): The world map to use (default is `KanjiDeGoMap`).
            speed (float, optional): The speed multiplier for bot actions (default is 1.0).
        """
        self.kadego: Page = kadego
        self.mode: str = mode
        self.level: str = level
        self.strategy: Strategy = strategy
        self.lives: int = lives
        self.questions: int = questions
        self.logger: Logger = logger or BaseLogger()
        self.world_map: WorldMap = world_map or KanjiDeGoMap()
        self.speed: float = speed

        self.navigator: MenuNavigator = MenuNavigator(
            self.kadego, self.world_map, self.speed
        )
        self.game_status: GameStatus = GameStatus(0, 0)
        self.status: str = "Initialized"

        self._validate()

    def _validate(self) -> None:
        """Check if the bot's configuration is valid."""
        if self.mode not in self.world_map.modes:
            raise InvalidBotConfigurationError(f"Invalid mode: {self.mode}")

        if self.level not in self.world_map.mode_levels.get(self.mode, []):
            raise InvalidBotConfigurationError(
                f"Invalid level: {self.level} for mode: {self.mode}"
            )

        if not (0 < self.lives < 6):
            raise InvalidBotConfigurationError(
                f"Invalid number of lives: {self.lives}. Must be between 1 and 5."
            )

        if self.questions not in (7, 10, 16):
            raise InvalidBotConfigurationError(
                f"Invalid number of questions: {self.questions}. Must be 7, 10, or 16."
            )

    def _get_new_game_status(self) -> GameStatus:
        """Returns a new game status object based on the specified lives and
        questions.

        Returns:
            GameStatus: A new game status object.
        """
        return GameStatus(self.lives, self.questions)

    async def _wait(self, delay: float) -> None:
        """Pauses the bot for a specified amount of time, scaled by the bot's
        speed.

        Args:
            delay (float): The time to wait in seconds.
        """
        await asyncio.sleep(delay * self.speed)

    async def _clear_hell(self) -> None:
        """Clears hell once, then restarts the game."""
        self.status = "Clearing 'Hell'"
        strategy_temp = self.strategy
        logger_temp = self.logger
        self.strategy = AlwaysCorrectStrategy()
        self.logger = BaseLogger()

        await self.navigator.start_round("main", "hell", self.lives, self.questions)
        await self.run()
        await self.reload()

        self.strategy = strategy_temp
        self.logger = logger_temp

    async def _navigate(self) -> None:
        """Navigates through the menu."""
        self.status = "Navigating menu"
        await self.navigator.start_round(
            self.mode, self.level, self.lives, self.questions
        )

    async def _wait_for_round_to_load(self) -> None:
        """Waits for some time for the round to load.

        Returns:
            bool: True if the round loaded in time, False if it　did not.
        """
        selector = "#_111_input"
        for _ in range(60):
            element = await self.kadego.query_selector(selector)
            if element:
                return
            await press_key(self.kadego, "Space")
            await self._wait(Speed.MEDIUM.value)
        raise TimeoutError

    async def _wait_for_question(self) -> None:
        """Waits for the next question."""
        self.status = "Waiting for question"
        selector = "#_111_input"
        await self.kadego.wait_for_selector(selector)

    async def _get_answer(self) -> tuple[str, QuestionOutcome]:
        """Retrieves an answer to the current question using the selected
        strategy.

        Returns:
            tuple[str, QuestionOutcome]: The answer and the　outcome of the question.
        """
        game_variables = await get_game_variables(self.kadego)
        answer, outcome = self.strategy.solve(game_variables, self.game_status)
        return answer, outcome

    async def _answer_question(self, answer: str, outcome: QuestionOutcome) -> None:
        """Provides an answer to the current question and updates the game
        status according to the provided outcome.

        Args:
            answer (str): The answer to submit.
            outcome　(QuestionOutcome): The outcome of the question.
        """
        self.status = f"Answering question with '{answer}'"
        selector = "#_111_input"
        await self.kadego.fill(selector, answer)
        await self.kadego.press(selector, "Enter")
        self.game_status.update(outcome)

    async def _wait_for_solution(self, outcome: QuestionOutcome) -> None:
        """Waits for the solution to be displayed and logs the current game
        state.

        Args:
            outcome (QuestionOutcome): The outcome of the question.
        """
        self.status = f"Waiting for solution"
        await self._wait(Speed.MEDIUM.value)
        game_variables = await get_game_variables(self.kadego)
        self.logger.log(game_variables)
        if not outcome.skipped and not outcome.correct:
            await press_key(self.kadego, "Space")

    async def _wait_for_round_reload_loop(self) -> None:
        """Waits for a new round to start, reloading the game if stuck."""
        self.status = "Waiting for round to load"
        self.game_status: GameStatus = self._get_new_game_status()
        await self._wait_for_round_to_load()

    async def _answer_questions_loop(self) -> None:
        """Waits for and answers questions in a loop until the game is
        finished."""
        while not self.game_status.is_done():
            await self._wait_for_question()
            answer, outcome = await self._get_answer()
            await self._answer_question(answer, outcome)
            await self._wait_for_solution(outcome)

    async def _wait_for_round_to_finish(self) -> None:
        """Waits a little at the end of a round for e.g. the boss defeated
        animation to finish."""
        self.status = "Waiting for game to finish"
        if self.game_status.is_win():
            await self._wait(Speed.LONG.value)
        else:
            await self._wait(Speed.SHORT.value)

    async def _retry(self) -> None:
        """Starts a new round by retrying."""
        self.status = "Retrying"
        keys = 2 * ["ArrowDown"]
        await press_key_sequence(self.kadego, keys, Speed.SHORT.value)
        self.game_status = self._get_new_game_status()

    async def launch(self) -> None:
        """Navigates through the game performing all necessary setup and
        finally starts a round."""
        if (self.mode, self.level, self.navigator.mode) == ("rush", "level_7", None):
            await self._clear_hell()
        await self._navigate()

    async def reload(self) -> None:
        """Reloads the game and starts a new round again."""
        await self.navigator.reload()
        await self.launch()

    async def run(self) -> None:
        """Runs the bot for one round of play."""
        await self._wait_for_round_reload_loop()
        await self._answer_questions_loop()
        await self._wait_for_round_to_finish()
        await self._retry()



# Standard Library Imports
from abc import ABC, abstractmethod


class Logger(ABC):
    """An Abstract base class for logging game variables.

    A `Logger` is used by a `Bot` to log game data. Each time the solution
    to a question is revealed the `Bot` will call the `Logger` to log the
    current game variables.
    """

    @abstractmethod
    def __init__(self) -> None:
        """Initializes the logger."""
        pass

    @abstractmethod
    def log(self, game_variables: dict) -> None:
        """Logs the given game variables.

        Args:
            game_variables (dict): The game variables shortly after a solution was revealed.
        """
        pass



# Local Imports
from .strategy import Strategy
from ..solvers.correct_solver import CorrectSolver
from ..solvers.solver import Solver
from ..utils.game_status import GameStatus


class AlwaysCorrectStrategy(Strategy):
    """A `Strategy` that always chooses the `CorrectSolver` to answer
    questions."""

    def get_solver(self, game_variables: dict, status: GameStatus) -> Solver:
        """Chooses the correct answer to the question.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.
            status (GameStatus): The current game status (e.g. number of lives)

        Returns:
            tuple[str, QuestionOutcome]: A tuple of the correct answer and the 'Correct' outcome.
        """
        return CorrectSolver()



Modes = dict[str, int]
ModeLevels = dict[str, dict[str, int]]
Start = list[str]
RoundStarts = dict[str, dict[str, Start]]
Popups = list[str]
ModePopups = dict[str, Popups]
ModeLevelPopups = dict[str, dict[str, Popups]]


class WorldMap:
    """A class representing a world map of a game.

    A `WorldMap` consists of different modes which in turn have levels. Each of
    these has a location which represents the amount of right arrow button
    presses it would take to get there. Additionally, the game in general, each
    mode and each level in a mode can have a number of popups which appear as
    well as different methods to enter them.
    """

    def __init__(
        self,
        modes: Modes,
        mode_levels: ModeLevels,
        popups: Popups | None = None,
        mode_popups: ModePopups | None = None,
        mode_level_popups: ModeLevelPopups | None = None,
        round_starts: RoundStarts | None = None,
    ) -> None:
        """Initializes the WorldMap.

        Args:
            modes (Modes): A dictionary specifying the mode locations.
            mode_levels (ModeLevels): A dictionary specifying the level locations.
            popups (Popups, optional): A list of popups (default is list()).
            mode_popups (ModePopups, optional): A dictionary specifying the mode popups (default is dict()).
            mode_level_popups (ModeLevelPopups, optional): A dictionary specifying the level popups (default is dict()).
            round_starts (RoundStart, optional): A dictionary specifying the round starts (default is dict()).
        """
        self.modes: Modes = modes
        self.mode_levels: ModeLevels = mode_levels
        self.popups: Popups = popups or []
        self.mode_popups: ModePopups = mode_popups or dict()
        self.mode_level_popups: ModeLevelPopups = mode_level_popups or dict()
        self.round_starts: RoundStarts = round_starts or dict()

    def get_modes(self) -> list[str]:
        """Retrieves the list of modes.

        Returns:
            list[str]: The list of modes.
        """
        return list(self.modes.keys())

    def get_levels(self, mode: str) -> list[str]:
        """Retrieves the list of levels of a specified mode.

        Args:
            mode (str): The mode of the requested levels.

        Returns:
            list[str]: The list of levels.
        """
        return list(self.mode_levels[mode].keys())

    def get_mode_direction(self, mode: str) -> int:
        """Retrieves the location of a mode.

        Args:
            mode (str): The mode for which the location is requested.

        Returns:
            int: The amount of right arrow button presses necessary to get to the location.
        """
        return self.modes[mode]

    def get_level_direction(self, mode: str, level: str) -> int:
        """Retrieves the location of a level in a mode.

        Args:
            mode (str): The mode of the given level.
            level (str): The level for which the location is requested.

        Returns:
            int: The amount of right arrow button presses necessary to get to the level.
        """
        return self.mode_levels[mode][level]

    def get_popups(self) -> Popups:
        """Gives instructions on how to close popups.

        Returns:
            Popups: The list of button presses necessary to close the popups.
        """
        return self.popups

    def get_mode_popups(self, mode: str) -> Popups:
        """Gives instructions on how to close popups in a mode.

        Args:
            mode (str): The mode in which the popups appear.

        Returns:
            Popups: The list of button presses necessary to close the popups.
        """
        return self.mode_popups.get(mode) or []

    def get_mode_level_popups(self, mode: str, level: str) -> Popups:
        """Gives instructions on how to close popups in a level of a mode.

        Args:
            mode (str): The mode of the given level.
            level (str): The level in which the popups appear.

        Returns:
            Popups: The list of button presses necessary to close the popups.
        """
        level_popups = self.mode_level_popups.get(mode)
        if level_popups is None:
            return []
        return level_popups.get(level) or []

    def clear_popups(self) -> None:
        """Removes the popups."""
        self.popups = []

    def clear_mode_popups(self, mode: str) -> None:
        """Removes the popups in a mode.

        Args:
            mode (str): The mode in which the popups appear.
        """
        if mode not in self.mode_popups:
            return
        self.mode_popups.pop(mode)

    def clear_mode_level_popups(self, mode: str, level: str) -> None:
        """Removes the popups in a level of a mode.

        Args:
            mode (str): The mode of the given level.
            level (str): The level in which the popups appear.
        """
        level_popups = self.mode_level_popups.get(mode)
        if level_popups is None or level not in level_popups:
            return
        level_popups.pop(level)

    def get_round_start(self, mode: str, level: str) -> Start:
        """Gives instructions on how to start a round for a level of a mode.

        Args:
            mode (str): The mode of the given level.
            level (str): The level to of which to start a round.

        Returns:
            Popups: The list of button presses necessary to start the round.
        """
        round_starts_level = self.round_starts.get(mode)
        if round_starts_level is None:
            return []
        return round_starts_level.get(level) or ["Space"]



# Standard Library Imports
import json
from abc import abstractmethod

# Local Imports
from .logger import Logger


class JsonFileLogger(Logger):
    """An Abstract base class for logging game variables into a JSON file.

    This `Logger` temporarily stores the data from the game variables and
    writes it into a JSON file.
    """

    def __init__(self, path: str) -> None:
        """Initializes the logger.

        Args:
            path (str): The path to the JSON file where log data will be stored.
        """
        super().__init__()
        self.path: str = path
        self.data: dict = self._read_data()

    def _read_data(self) -> dict:
        """Reads data from the specified JSON file.

        Returns:
            dict: The data read from the JSON file.
        """
        try:
            with open(self.path, "r") as file:
                return json.load(file)
        except FileNotFoundError:
            return dict()

    def _write_data(self) -> None:
        """Writes the current log data to the specified JSON file."""
        data_lines = [
            f'  "{key}": {json.dumps(value, ensure_ascii=False)}'
            for key, value in self.data.items()
        ]
        data_string = "{\n" + ",\n".join(data_lines) + "\n}"

        with open(self.path, "w", encoding="utf-8") as file:
            file.write(data_string)

    @abstractmethod
    def log(self, game_variables: dict) -> None:
        """Logs the given game variables to be written into a JSON file.

        Args:
            game_variables (dict): The game variables shortly after a solution was revealed.
        """
        pass



# Standard Library Imports
import asyncio
from abc import ABC, abstractmethod

# Local Imports
from ..bots.bot import Bot


class Observer(ABC):
    """An Abstract base class for observing a bot.

    An `Observer` runs in parallel to a `Bot` reacting to changes in its
    state or outside factors e.g. by stopping the `Runner` from running
    the `Bot`.
    """

    def __init__(self, bot: Bot):
        """Initializes the observer given a bot.

        Args:
            bot (Bot): The bot to be observed.
        """
        self.bot: Bot = bot

    @abstractmethod
    async def observe(self, stop_event: asyncio.Event) -> None:
        """Observes the bot reacting to changes or outside factors.

        Args:
            stop_event (asyncio.Event): Event that when set stops the runner from running the bot.
        """
        pass



# Standard Library Imports
import asyncio

# Local Imports
from .periodic_observer import PeriodicObserver


class StatusObserver(PeriodicObserver):
    """An `Observer` priting the current status of a `Bot`.

    This `Observer` prints the current state of a `Bot` periodically
    overwriting it.
    """

    def _loop_function(self, stop_event: asyncio.Event) -> None:
        """Prints the current state of the bot.

        Args:
            stop_event (asyncio.Event): Event that when set stops the runner from running the bot.
        """
        status = self.bot.status
        print(f"\rStatus: {status}\033[K", end='', flush=True)



# Standard Library Imports
import asyncio
from typing import Sequence

# Third-Party Imports
from playwright.async_api import Page, async_playwright

GAME_VARIABLES = "window.$gameVariables"
KEY_PRESS_HOLD_DOWN = .1
TIMEOUT = 100000
URL = "https://html5.plicy.net/GamePlay/155561"


async def _launch_browser(headless: bool) -> Page:
    """Launches a Chromium browser instance using Playwright and returns a
    page.

    Args:
        headless (bool): Whether to launch the browser in headless mode (without UI).

    Returns:
        Page: A Playwright Page object representing the newly opened browser page.
    """
    browser = await async_playwright().start()
    context = await browser.chromium.launch(headless=headless)
    page = await context.new_page()
    page.set_default_timeout(TIMEOUT)
    return page


async def get_kadego(headless: bool = True) -> Page:
    """Opens 漢字でGO! in a new browser instance and returns the page.

    Args:
        headless (bool, optional): Whether to run the browser in headless mode (default is True).

    Returns:
        Page: A Playwright Page object representing the page after opening 漢字でGO!.
    """
    page = await _launch_browser(headless=headless)
    await page.goto(URL)
    return page


async def press_key(page: Page, key: str) -> None:
    """Simulates pressing a key on the keyboard.

    Args:
        page (Page): The Playwright Page object.
        key (str): The key to press.
    """
    keys = key.split("+")
    for k in keys:
        await page.keyboard.down(k)
        await asyncio.sleep(KEY_PRESS_HOLD_DOWN)
    for k in keys[::-1]:
        await page.keyboard.up(k)


async def press_key_sequence(page: Page, keys: Sequence[str], delay: float) -> None:
    """Simulates pressing a sequence of keys on the keyboard.

    Args:
        page (Page): The Playwright Page object.
        keys (Sequence[str]): The sequence of keys to press.
        delay (float): The delay (in seconds) between key presses
    """
    for i, key in enumerate(keys):
        await press_key(page, key)
        if i != len(keys) - 1:
            await asyncio.sleep(delay)


async def get_game_variables(page: Page) -> dict:
    """Retrieves the current game variables from a page.

    Args:
        page (Page): The Playwright Page object.

    Returns:
        game_variables (dict): The current game variables.
    """
    game_variables = await page.evaluate(GAME_VARIABLES)
    return game_variables



# Standard Library Imports
import asyncio

# Local Imports
from ..bots.bot import Bot
from ..observers.observer import Observer


class Runner:
    """A class responsible for running a bot playing 漢字でGO!

    A `Runner` can launch a `Bot` and an `Observer` to monitor the bot's
    activity.
    """

    def __init__(self, bot: Bot, observer: Observer) -> None:
        """Initializes the runner with a bot and an optional observer.

        Args:
            bot (Bot): The bot that will be launched and executed by the Runner.
            observer (Observer): The observer for the bot.
        """
        self.bot: Bot = bot
        self.observer: Observer = observer
        self.stop_event: asyncio.Event = asyncio.Event()

    async def run(self) -> None:
        """Starts running the bot in a loop as well as the observer.

        This method launches the bot and starts monitoring the bot's
        activity. The bot will continue running in a loop until the
        observer sends a signal to stop.
        """
        asyncio.create_task(self.observer.observe(self.stop_event))
        await self.bot.launch()

        while not self.stop_event.is_set():
            try:
                await self.bot.run()
            except TimeoutError:
                await self.bot.reload()



# Local Imports
from .solver import Solver
from ..utils.question_outcome import QuestionOutcome


class CorrectSolver(Solver):
    """A `Solver` that always chooses the correct answer."""

    def solve(self, game_variables: dict) -> tuple[str, QuestionOutcome]:
        """Chooses the correct answer to the question.

        Args:
            game_variables (dict): The game variables shortly after a question was posed.

        Returns:
            tuple[str, QuestionOutcome]: A tuple of the correct answer and the 'Correct' outcome.
        """
        return game_variables["_data"][9], QuestionOutcome(True, False)



# Third-Party Imports
from playwright.async_api import Page

# Local Imports
from .runner import Runner
from ..bots.bot import Bot
from ..loggers.single_level_words_logger import SingleLevelWordsLogger
from ..observers.progress_bar_observer import ProgressBarObserver
from ..strategies.curious_strategy import CuriousStrategy


class SingleLevelScraper(Runner):
    """A runner for scraping the word data of individual levels from 漢字でGO!.

    This class sets up a `Bot` configured to play the 'Rush' mode with the
    `CuriousStrategy` as well as a `SingleLevelWordsLogger` to log the
    scraped words and a `ProgressBarObserver` to display the progress.
    """

    def __init__(self, kadego: Page, level: int, path: str) -> None:
        """Initializes the runner with all its components.

        Args:
            kadego (Page): The Playwright page for playing 漢字でGO!.
            level (int): The level to be scraped for words.
            path (str): The file path where the scraped data will be stored.
        """
        logger = SingleLevelWordsLogger(level, path)
        strategy = CuriousStrategy()

        bot = Bot(kadego, "rush", f"level_{level}", strategy, logger=logger)
        observer = ProgressBarObserver(bot)

        super().__init__(bot, observer)



# Standard Library Imports
import asyncio

# Third-Party Imports
from tqdm import tqdm

# Local Imports
from .periodic_observer import PeriodicObserver
from ..bots import Bot
from ..loggers import SingleLevelWordsLogger


class ProgressBarObserver(PeriodicObserver):
    """An `Observer` priting the current status of a `Bot` as well as the
    progress via a progress bar.

    This `Observer` is meant to work together with a `Bot` that has an
    instance of `SingleLevelWordsLogger` as its `Logger`. A progress bar
    is printed that shows how many distinct words' data was already
    collected and updates periodically. Once all words' data hase been
    collected the `Runner` will be stopped from running the `Bot`.
    """

    def __init__(self, bot: Bot, delay: float = 0.1) -> None:
        """Initializes the observer given a bot and delay.

        Args:
            bot (Bot): The bot to be observed.
            delay (float, optional): The delay (in seconds) between each observation cycle (default is 0.2).
        """
        super().__init__(bot, delay)
        self.progress_bar: tqdm = tqdm(total=None)

    def _loop_function(self, stop_event: asyncio.Event) -> None:
        """Updates the progress bar based on the bot's logger data and stops
        the `Runner` from running the `Bot` once all words' data has been
        collected.

        Args:
            stop_event (asyncio.Event): Event that when set stops the runner from running the bot.
        """
        if isinstance(self.bot.logger, SingleLevelWordsLogger):
            data = self.bot.logger.data
            self.progress_bar.total = len(data)
            values = data.values()
            non_nones = sum(value is not None for value in values)
        else:
            non_nones = 0

        self.progress_bar.set_description(self.bot.status[:50].ljust(50))
        self.progress_bar.n = non_nones
        self.progress_bar.refresh()

        if non_nones == self.progress_bar.total:
            stop_event.set()
            self.progress_bar.close()



# Standard Library Imports
import asyncio
from abc import abstractmethod

# Local Imports
from .observer import Observer
from ..bots import Bot


class PeriodicObserver(Observer):
    """An abstract base class where the observation is performed periodically.

    This `Observer` performs the observation in a loop periodically waiting
    for a specified amount of time.
    """

    def __init__(self, bot: Bot, delay: float = 0.1) -> None:
        """Initializes the observer given a bot and delay.

        Args:
            bot (Bot): The bot to be observed.
            delay (float, optional): The delay (in seconds) between each observation cycle (default is 0.2).
        """
        super().__init__(bot)
        self.delay: float = delay

    @abstractmethod
    def _loop_function(self, stop_event: asyncio.Event) -> None:
        """Performs the actual observation

        Args:
            stop_event (asyncio.Event): Event that when set stops the runner from running the bot.
        """
        pass

    async def observe(self, stop_event: asyncio.Event) -> None:
        """Runs the function performing the actual observation in a loop.

        Args:
            stop_event (asyncio.Event): Event that when set stops the runner from running the bot.
        """
        while not stop_event.is_set():
            self._loop_function(stop_event)
            await asyncio.sleep(self.delay)
